/*
 * Copyright 2011 Red Hat, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "codec_json.h"

#include "asn1/asn_SET_OF.h"
#include "asn1/constr_SET_OF.h"

#include "asn1/OTPFlags.h"
#include "asn1/AnyURI.h"
#include "asn1/Int32.h"
#include "asn1/UInt32.h"
#include "asn1/OTPFormat.h"
#include "asn1/OBJECT_IDENTIFIER.h"
#include "asn1/KerberosTime.h"

#include "memory.h"

#include <time.h>

typedef json_object *
(jsonEncoder)(const void *ptr);
typedef bool
(jsonDecoder)(void *ptr, const json_object *obj);

#define CODEC(type, enctype, dectype) \
  { &asn_DEF_ ## type, sizeof(type ## _t), \
    (jsonEncoder*) enctype ## _encode, \
    (jsonDecoder*) dectype ## _decode }

/* Base64 adapted from http://base64.sourceforge.net/b64.c under MIT license */
static const char cb64[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdef"
                           "ghijklmnopqrstuvwxyz0123456789+/";
static const char cd64[] = "|$$$}rstuvwxyz{$$$$$$$>?@ABCDEFGHIJKLMNOP"
                           "QRSTUVW$$$$$$XYZ[\\]^_`abcdefghijklmnopq";

static void
b64_encode_block(unsigned char in[3], unsigned char out[4], int len)
{
  out[0] = cb64[in[0] >> 2];
  out[1] = cb64[((in[0] & 0x03) << 4) | ((in[1] & 0xf0) >> 4)];
  out[2] = (unsigned char) (len > 1 ? cb64[((in[1] & 0x0f) << 2)
                                          |((in[2] & 0xc0) >> 6)]
                                    : '=');
  out[3] = (unsigned char) (len > 2 ? cb64[in[2] & 0x3f] : '=');
}

static void
b64_decode_block(unsigned char in[4], unsigned char out[3])
{
  out[0] = (unsigned char) (in[0] << 2 | in[1] >> 4);
  out[1] = (unsigned char) (in[1] << 4 | in[2] >> 2);
  out[2] = (unsigned char) (((in[2] << 6) & 0xc0) | in[3]);
}
/* End Base64 adaptation from http://base64.sourceforge.net/b64.c under MIT */

static json_object *
BIT_STRING_encode(const BIT_STRING_t *ptr)
{
  int flags = 0, i;

  if (ptr->size > sizeof(uint32_t))
    return false;

  for (i = 0; i < ptr->size * 8 - ptr->bits_unused; i++)
    flags |= ptr->buf[i / 8] & (1 << (7 - i % 8)) ? 1 << i : 0;

  return json_object_new_int(flags);
}

static bool
BIT_STRING_decode(BIT_STRING_t *ptr, json_object *obj)
{
  int flags, i;

  if (!json_object_is_type(obj, json_type_int))
    return false;

  ptr->buf = malloc0(sizeof(uint32_t));
  if (!ptr->buf)
    return false;
  ptr->size = sizeof(uint32_t);

  flags = json_object_get_int(obj);
  for (i = 0; i < sizeof(uint32_t) * 8 - ptr->bits_unused; i++)
    ptr->buf[i / 8] |= flags & (1 << i) ? (1 << (7 - i % 8)) : 0;

  return true;
}

static json_object *
OCTET_STRING_encode(const OCTET_STRING_t *ptr)
{
  unsigned char in[3], out[4], *str = NULL;
  json_object *tmp;
  size_t i;

  str = malloc(sizeof(char) * (ptr->size / 3 * 4 + ptr->size % 3 + 4));
  if (!str)
    return NULL;
  str[0] = '\0';

  for (i=0 ; i < ptr->size ; i++) {
    if (i % 3 == 0)
      memset(in, 0, sizeof(in));

    in[i % 3] = ptr->buf[i];

    if (i % 3 == 2 || i + 1 == ptr->size) {
      b64_encode_block(in, out, i % 3 + 1);
      strncat((char*) str, (char*) out, 4);
    }
  }

  tmp = json_object_new_string((char*) str);
  free(str);
  return tmp;
}

static bool
OCTET_STRING_decode(OCTET_STRING_t *ptr, json_object *obj)
{
  unsigned char in[4], out[3], c;
  const char *str;

  if (!json_object_is_type(obj, json_type_string))
    return false;

  str = json_object_get_string(obj);
  if (!str)
    return false;

  ptr->buf = malloc(sizeof(char) * (strlen(str) / 4 * 3 + 3 + 1));
  if (!ptr->buf)
    return false;

  /* Adapted from http://base64.sourceforge.net/b64.c under MIT license */
  for (int i = 0, j = 0; str[i]; i++) {
    if (j % 4 == 0)
      memset(in, 0, sizeof(in));

    c = (unsigned char) str[i];
    c = (unsigned char) ((c < 43 || c > 122) ? 0 : cd64[c - 43]);
    c = (unsigned char) c ? ((c == '$') ? 0 : c - 61) : 0;
    if (str[i+1] && c == 0)
      continue;

    in[j % 4] = (unsigned char) c - 1;
    if (j++ % 4 == 3 || !str[i+1]) {
      b64_decode_block(in, out);
      strncat((char*) ptr->buf, (char*) out, (j - 1) % 4);
    }
  }

  ptr->size = (int) strlen((char*) ptr->buf);
  return true;
}

static json_object *
UTF8String_encode(const UTF8String_t *ptr)
{
  return json_object_new_string_len((char*) ptr->buf, ptr->size);
}

static bool
UTF8String_decode(UTF8String_t *ptr, json_object *obj)
{
  if (!json_object_is_type(obj, json_type_string))
    return false;

  ptr->buf = (unsigned char*) json_object_get_string(obj);
  if (ptr->buf)
    ptr->buf = (unsigned char*) strdup((char*) ptr->buf);
  if (ptr->buf)
    ptr->size = (int) strlen((char*) ptr->buf);

  return true;
}

static json_object *
Int32_encode(const Int32_t *ptr)
{
  return json_object_new_int(*ptr);
}

static bool
Int32_decode(Int32_t *ptr, json_object *obj)
{
  if (!json_object_is_type(obj, json_type_int))
    return false;

  *ptr = json_object_get_int(obj);
  return true;
}

static json_object *
UInt32_encode(const UInt32_t *ptr)
{
  long int i;

  if (asn_INTEGER2long(ptr, &i) != 0)
    return false;

  return json_object_new_int(i);
}

static bool
UInt32_decode(UInt32_t *ptr, json_object *obj)
{
  if (!json_object_is_type(obj, json_type_int))
    return false;

  return asn_long2INTEGER(ptr, json_object_get_int(obj)) == 0;
}

static json_object *
INTEGER_encode(const INTEGER_t *ptr)
{
  long int i;

  if (asn_INTEGER2long(ptr, &i) < 0)
    return false;

  return json_object_new_int(i);
}

static bool
INTEGER_decode(INTEGER_t *ptr, json_object *obj)
{
  if (!json_object_is_type(obj, json_type_int))
    return false;

  if (asn_long2INTEGER(ptr, json_object_get_int(obj)) < 0)
    return false;

  return true;
}

static bool
OTPFormat_decode(OTPFormat_t *ptr, json_object *obj)
{
  int value = 0;

  if (!json_object_is_type(obj, json_type_int))
    return false;

  value = json_object_get_int(obj);
  if (value < OTPFormat_decimal || value > OTPFormat_base64)
    return false;

  return INTEGER_decode(ptr, obj);
}

static json_object *
OBJECT_IDENTIFIER_encode(const OBJECT_IDENTIFIER_t *ptr)
{
  char *tmp = NULL, *out = NULL;
  long *arcs = NULL;
  int slots = 0, i;

  slots = OBJECT_IDENTIFIER_get_arcs((OBJECT_IDENTIFIER_t*) ptr,
                                     arcs, sizeof(long), 0);
  if (slots >= 0) {
    arcs = calloc0(slots, sizeof(long));
    if (!arcs)
      return false;
    slots = OBJECT_IDENTIFIER_get_arcs((OBJECT_IDENTIFIER_t*) ptr,
                                       arcs, sizeof(long), slots);
  }
  if (slots < 0) {
    free(arcs);
    return false;
  }

  for (i = 0; i < slots; i++) {
    if (asprintf(&tmp, "%s.%d", out ? out : "", arcs[i]) < 0) {
      free(arcs);
      free(out);
      out = NULL;
      return false;
    }

    free(out);
    out = tmp;
  }

  memmove(out, out + 1, strlen(out)); /* Remove the leading . */
  free(arcs);

  return json_object_new_string(out);
}

static bool
OBJECT_IDENTIFIER_decode(OBJECT_IDENTIFIER_t *ptr, json_object *obj)
{
  const char *tmp = NULL;
  long *arcs = NULL;
  int slots = 0;

  if (!json_object_is_type(obj, json_type_string)
      || !(tmp = json_object_get_string(obj)))
    return false;

  slots = OBJECT_IDENTIFIER_parse_arcs(tmp, -1, arcs, 0, NULL);
  if (slots >= 0) {
    arcs = calloc0(slots, sizeof(long));
    if (!arcs)
      return false;
    slots = OBJECT_IDENTIFIER_parse_arcs(tmp, -1, arcs, slots, NULL);
  }
  if (slots < 0) {
    free(arcs);
    return false;
  }

  if (OBJECT_IDENTIFIER_set_arcs(ptr, arcs, sizeof(long), slots) < 0) {
    free(arcs);
    return false;
  }

  return true;
}

static json_object *
GeneralizedTime_encode(const GeneralizedTime_t *ptr)
{
  struct tm zero(time);
  time_t tmp;

  if (asn_GT2time(ptr, &time, 0) < 0)
    return false;

  tmp = mktime(&time);
  if (tmp < 0)
    return false;

  return json_object_new_int(tmp);
}

static bool
GeneralizedTime_decode(GeneralizedTime_t *ptr, json_object *obj)
{
  time_t time;

  if (!json_object_is_type(obj, json_type_int))
    return false;

  time = json_object_get_int(obj);
  return asn_time2GT(ptr, gmtime(&time), 0);
}

static struct
{
  asn_TYPE_descriptor_t *td;
  unsigned int size;
  jsonEncoder *encode;
  jsonDecoder *decode;
} json_codec[] = {
  CODEC(BIT_STRING,        BIT_STRING,        BIT_STRING),
  CODEC(KerberosFlags,     BIT_STRING,        BIT_STRING),
  CODEC(OTPFlags,          BIT_STRING,        BIT_STRING),
  CODEC(OCTET_STRING,      OCTET_STRING,      OCTET_STRING),
  CODEC(UTF8String,        UTF8String,        UTF8String),
  CODEC(AnyURI,            UTF8String,        UTF8String),
  CODEC(Int32,             Int32,             Int32),
  CODEC(UInt32,            UInt32,            UInt32),
  CODEC(INTEGER,           INTEGER,           INTEGER),
  CODEC(OTPFormat,         INTEGER,           OTPFormat),
  CODEC(OBJECT_IDENTIFIER, OBJECT_IDENTIFIER, OBJECT_IDENTIFIER),
  CODEC(GeneralizedTime,   GeneralizedTime,   GeneralizedTime),
  CODEC(KerberosTime,      GeneralizedTime,   GeneralizedTime),
  { NULL, 0, NULL, NULL }
};

bool
int_json_decode(asn_TYPE_descriptor_t *type_descriptor, void *ptr,
                const char *buf, int len)
{
  json_tokener *tok = NULL;
  json_object *obj = NULL;
  bool result;

  if (!buf)
    return false;

  if (len < 0)
    len = strlen(buf);

  /* Parse our buffer into a json object */
  tok = json_tokener_new();
  if (tok) {
    obj = json_tokener_parse_ex(tok, buf, len);
    json_tokener_free(tok);
  }
  if (!obj)
    return NULL;

  /* Decode */
  result = int_json_decode_object(type_descriptor, ptr, obj);
  json_object_put(obj);
  return result;
}

static unsigned int
get_struct_size(asn_TYPE_descriptor_t *type_descriptor)
{
  /* Both asn_SEQUENCE_specifics_t and asn_SET_OF_specifics_t
   * start with 'int struct_size'. Yes, this is a hack. */
  struct {
    int struct_size;
  } *specs = type_descriptor->specifics;
  int i;

  if (specs)
    return specs->struct_size;

  for (i = 0; json_codec[i].td; i++)
    if (json_codec[i].td == type_descriptor)
      return json_codec[i].size;

  return 0;
}

bool
int_json_decode_object(asn_TYPE_descriptor_t *type_descriptor, void *ptr,
                       json_object *obj)
{
  int i;

  /* Simple type */
  if (type_descriptor->elements_count == 0) {
    for (i = 0; json_codec[i].td; i++) {
      if (json_codec[i].td == type_descriptor)
        return json_codec[i].decode(ptr, obj);
    }

    return false;
  }

  /* SET OF or SEQUENCE OF with only one child type */
  if (type_descriptor->free_struct == SET_OF_free
      && type_descriptor->elements_count == 1) {
    struct {
      A_SET_OF(void) list;
      asn_struct_ctx_t _asn_ctx;
    } *set = ptr;
    unsigned int size = get_struct_size(type_descriptor->elements[0].type);

    if (!json_object_is_type(obj, json_type_array) || !size)
      return false;

    set->list.size = json_object_array_length(obj) * sizeof(void*);
    set->list.array = malloc0(set->list.size);
    if (!set->list.array)
      return false;

    for (i = 0; i < set->list.size / sizeof(void*); i++) {
      json_object *item = json_object_array_get_idx(obj, i);
      if (!item)
        return false;

      set->list.array[i] = malloc0(size);
      if (!set->list.array[i])
        return false;

      if (!int_json_decode_object(type_descriptor->elements[0].type,
          set->list.array[i], item))
        return false;

      set->list.count++;
    }
  }

  /* SEQUENCE type */
  else if (type_descriptor->free_struct != SET_OF_free) {
    if (!json_object_is_type(obj, json_type_object))
      return false;

    for (i = 0; i < type_descriptor->elements_count; i++) {
      struct asn_TYPE_member_s *e;
      json_object *value = NULL;
      void *eptr;

      e = &type_descriptor->elements[i];
      eptr = ptr + e->memb_offset;
      if (!eptr)
        return false;

      value = json_object_object_get(obj, e->name);

      if (e->flags & ATF_POINTER) {
        void **pp = (void**) eptr;

        if (!*pp && value) {
          if (get_struct_size(e->type))
            *pp = malloc0(get_struct_size(e->type));
          if (!*pp)
            return false;
        }

        eptr = *pp;
      }

      if (value && !int_json_decode_object(e->type, eptr, value))
        return false;
    }
  }

  if (type_descriptor->check_constraints(type_descriptor, ptr, NULL, NULL) != 0)
    return false;

  return true;
}

char *
int_json_encode(asn_TYPE_descriptor_t *type_descriptor, const void *ptr,
                unsigned int *len)
{
  const char *tmp;
  json_object *obj;

  obj = int_json_encode_object(type_descriptor, ptr);
  if (!obj)
    return NULL;

  tmp = json_object_to_json_string(obj);
  if (len)
    *len = strlen(tmp);
  return strdup(tmp);
}

json_object *
int_json_encode_object(asn_TYPE_descriptor_t *type_descriptor, const void *ptr)
{
  int i;

  /* Simple type */
  if (type_descriptor->elements_count == 0) {
    for (i = 0; json_codec[i].td; i++)
      if (json_codec[i].td == type_descriptor)
        return json_codec[i].encode(ptr);
    return NULL;
  }

  /* SET OF or SEQUENCE OF with only one child type */
  else if (type_descriptor->free_struct == SET_OF_free
      && type_descriptor->elements_count == 1) {
    const struct set_struct
    {
      A_SET_OF(void) list;
      asn_struct_ctx_t _asn_ctx;
    }*set = ptr;
    json_object *array;

    array = json_object_new_array();
    if (!array)
      return false;

    for (i = 0; i < set->list.count; i++) {
      json_object *tmp;

      tmp = int_json_encode_object(type_descriptor->elements[0].type,
          set->list.array[i]);
      if (!tmp) {
        json_object_put(array);
        return false;
      }

      json_object_array_add(array, tmp);
    }

    return array;
  }

  /* SEQUENCE type */
  else if (type_descriptor->free_struct != SET_OF_free) {
    json_object *obj;

    obj = json_object_new_object();
    if (!obj)
      return false;

    for (i = 0; i < type_descriptor->elements_count; i++) {
      struct asn_TYPE_member_s *e;
      json_object *tmp;
      const void *eptr;

      e = &type_descriptor->elements[i];
      eptr = ptr + e->memb_offset;

      if (eptr && e->flags & ATF_POINTER)
        eptr = *((void**) eptr);

      if (!eptr)
        continue;

      tmp = int_json_encode_object(e->type, eptr);
      if (!tmp) {
        json_object_put(obj);
        return false;
      }

      json_object_object_add(obj, e->name, tmp);
    }

    return obj;
  }

  return NULL;
}
